
/**************************************************************************
 *
 *  $Id: mbg_ntp_test_util.h 1.7 2018/11/26 11:59:31Z martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Definitions and prototypes used with mbg_ntp_test_util.c
 *
 * -----------------------------------------------------------------------
 *  $Log: mbg_ntp_test_util.h $
 *  Revision 1.7  2018/11/26 11:59:31Z  martin
 *  Moved some inline functions to timeutil.h.
 *  Revision 1.6  2018/03/09 09:55:33  martin
 *  Updated function prototypes.
 *  Revision 1.5  2016/11/21 17:22:50  martin
 *  Adjustable client poll value sent to server, defaults to 6.
 *  Revision 1.4  2016/08/05 12:37:31  martin
 *  New variable client_precision.
 *  Updated function prototypes.
 *  Revision 1.3  2016/07/14 10:55:10Z  martin
 *  Fixed build under Windows.
 *  Revision 1.2  2014/09/24 11:12:20Z  martin
 *  New structure NTP_RESULTS.
 *  Account for changed NTP_TSTAMP field names.
 *  Added and updated doxygen comments.
 *  Updated function prototypes.
 *  Revision 1.1  2014/04/01 13:02:21  martin
 *  Initial revision.
 *
 **************************************************************************/

#ifndef _MBG_NTP_TEST_UTIL_H
#define _MBG_NTP_TEST_UTIL_H


/* Other headers to be included */

#include <mbg_tgt.h>
#include <mbgntp.h>

#include <time.h>


#if defined( MBG_TGT_POSIX )
  #include <syslog.h>
  #include <arpa/inet.h>
#endif


#ifdef _MBG_NTP_TEST_UTIL
 #define _ext
 #define _DO_INIT
#else
 #define _ext extern
#endif



#if defined( _USE_PACK )
// not used here
// #pragma pack( 1 )      // set byte alignment
// #define _USING_BYTE_ALIGNMENT
#endif

/* Start of header body */

#ifdef __cplusplus
extern "C" {
#endif

#define INDENT   "  "

#if !defined( MBG_TGT_POSIX )
  // LOG_WARNING and friends are syslog stuff which
  // is not supported e.g. under Windows.
  enum
  {
    LOG_WARNING,
    LOG_ERR
  };
#endif

void mbglog( int priority, const char *fmt, ... );

_ext const char *newline
#ifdef _DO_INIT
 = "\n"
#endif
;


_ext int8_t client_precision
#ifdef _DO_INIT
 = -18
#endif
;


#define DEFAULT_CLIENT_POLL_VALUE  6

_ext uint8_t client_poll_value
#ifdef _DO_INIT
 = DEFAULT_CLIENT_POLL_VALUE
#endif
;


typedef struct
{
  double min;
  double max;

} MIN_MAX_DOUBLE;



// query mode statistics

typedef struct
{
  unsigned long long requests;
  unsigned long long replies;
  unsigned long long timeouts;

  MIN_MAX_DOUBLE turnaround;
  MIN_MAX_DOUBLE latency;
  MIN_MAX_DOUBLE ntp_delay;
  MIN_MAX_DOUBLE ntp_offs;

} QUERY_STATS;



typedef struct
{
  NTP_TSTAMP t_rcv;   ///< packet receive time in NTP format
  double turnaround;
  double latency;
  double ntp_delay;
  double ntp_offs;
  double t34;
  double t21;
  int auth_code;      ///< see ::MBG_NTP_AUTH_CODES

} NTP_RESULTS;



static __mbg_inline /*HDR*/
void update_min_max_double( double curr, MIN_MAX_DOUBLE *p )
{
  if ( curr > p->max )
    p->max = curr;

  if ( curr < p->min )
    p->min = curr;

}  // update_min_max_double



static __mbg_inline /*HDR*/
int check_bounds( int opt_val, int min_val, int max_val, int default_val )
{
  return ( ( opt_val < min_val ) || ( opt_val > max_val ) ) ? default_val : opt_val;

}  // check_bounds




/* function prototypes: */

/* ----- function prototypes begin ----- */

/* This section was generated automatically */
/* by MAKEHDR, do not remove the comments. */

 /**
 * @brief Reset query statistics
 *
 * @param[out] p  Statistics structure to be cleared.
 */
 void reset_query_stats( QUERY_STATS *p ) ;

 /**
 * @brief Print a statistics value of type "double"
 *
 * @param[in] info        Optional text string printed at beginning of line, or NULL.
 * @param[in] info_width  Width of the info text field.
 * @param[in] curr        Current statistics parameter.
 * @param[in] curr_width  With of the printed field containing the current parameter value.
 * @param[in] scale       Scale for the parameter values before they are printed.
 * @param[in] p           Optional minimum and maximum values of this parameter, or NULL.
 * @param[in] hint        Optional hint text appended at the end of the line, or NULL.
 */
 void print_stats_double( const char *info, int info_width, double curr, int curr_width, double scale, const MIN_MAX_DOUBLE *p, const char *hint ) ;

 /**
 * @brief Print a "struct timespec" as calendar date and time
 *
 * @param[in] p_ts  The time stamp to be printed, in struct timespec format.
 */
 void print_timespec_date_time( const struct timespec *p_ts ) ;

 /**
 * @brief Print date and time of an NTP time stamp
 *
 * @param[in] t_ntp  The time stamp to be printed, in NTP format.
 */
 void print_ntp_date_time( const NTP_TSTAMP *t_ntp ) ;

 /**
 * @brief Print a raw NTP timestamp, and optionally the derived date and time
 *
 * @param[in] msg  Text string printed at the beginning of the line.
 * @param[in] t    Time stamp, in NTP format.
 * @param[in] tail Optional text appended to the output, or NULL.
 * @param[in] print_date_time  If not 0, print also the derived date and time.
 */
 void print_ntp_time( const char *msg, const NTP_TSTAMP *t, const char *tail, int print_date_time ) ;

 /**
 * @brief Print contents of an NTP packet on several lines
 *
 * @param[in] msg  Header text to be printed on first line.
 * @param[in] p    An ::NTP_PACKET_INFO structure containing the NTP packet plus some additional information.
 * @param[in] print_date_time  If not 0, print also the date and time derived from the NTP time stamps.
 */
 void print_ntp_packet( const char *msg, const NTP_PACKET_INFO *p, int print_date_time, int verbose ) ;

 /**
 * @brief Initialize a basic NTP packet
 *
 * @param[in] p             The ::NTP_PACKET to be initialized.
 * @param[in] req_mode      Initalizer for the packet's mode field, e.g. ::MODE_CLIENT.
 * @param[in] prot_version  Initalizer for the packet's version field, e.g. ::DEFAULT_REQ_NTP_VERSION.
 */
 void init_ntp_req_packet( NTP_PACKET *p, int req_mode, int prot_version ) ;

 /**
 * @brief Evaluate the time stamps from an NTP packet received from a server
 *
 * @note The packet must already have been converted to host byte order.
 *
 * @param[out] p_rslt  ::NTP_RESULTS structure to take up the results of the computations.
 * @param[in]  p_rcv   The ::NTP_PACKET_INFO containing the packet received from the server, in host byte order.
 *
 * @see ::eval_ntp_packet
 * @see ::update_query_stats
 */
 void get_ntp_results( NTP_RESULTS *p_rslt, const NTP_PACKET_INFO *p_rcv ) ;

 /**
 * @brief Update query statistics with the results from the last query
 *
 * @param[in,out] p_stats      A ::QUERY_STATS structure to be updated.
 * @param[in]     p_rslt       ::NTP_RESULTS structure eventually set up by ::get_ntp_results.
 * @param[in]     new_queries  number of queries sent to yield the latest results (usually 1).
 *
 * @see ::get_ntp_results
 */
 void update_query_stats( QUERY_STATS *p_stats, const NTP_RESULTS *p_rslt, int new_queries ) ;

 /**
 * @brief Print query statistics
 *
 * @param[in] p_rslt      ::NTP_RESULTS structure eventually set up by ::get_ntp_results.
 * @param[in] p_stats     Optional ::QUERY_STATS structure, or NULL.
 * @param[in] info_width  Field with of the info strings.
 * @param[in] val_width   Field width for the values.
 * @param[in] scale       Scale for the values.
 */
 void print_query_stats( const NTP_RESULTS *p_rslt, const QUERY_STATS *p_stats, int info_width, int val_width, double scale ) ;

 /**
 * @brief Evaluate an NTP packet received from a server
 *
 * @note The packet to be evaluated must be in network byte order
 * and is converted to host byte order.
 *
 * @param[out]     p_rslt   Optional ::NTP_RESULTS structure to take up the results of the computations, or NULL.
 * @param[in,out]  p_rcv    The ::NTP_PACKET to be evaluated, in network byte order.
 * @param[in,out]  p_stats  A ::QUERY_STATS structure to be updated.
 *
 * @see ::ntoh_ntp_packet
 * @see ::get_ntp_results
 * @see ::update_query_stats
 */
 int eval_ntp_packet( NTP_RESULTS *p_rslt, NTP_PACKET_INFO *p_rcv, QUERY_STATS *p_stats ) ;

 /**
 * @brief Eval NTP packet and update statistics without returning results
 *
 * @param[in,out]  p_rcv    The ::NTP_PACKET to be evaluated, in network byte order.
 * @param[in,out]  p_stats  A ::QUERY_STATS structure to be updated.
 *
 * @see ::eval_ntp_packet
 */
 int update_ntp_stats( NTP_PACKET_INFO *p_rcv, QUERY_STATS *p_stats ) ;

 /**
 * @brief Print NTP results
 *
 * @param[in]  p_rslt   ::NTP_RESULTS structure eventually set up by ::eval_ntp_packet, or NULL.
 * @param[in]  p_req    The ::NTP_PACKET_INFO containing the request packet sent to the server, in host byte order, or NULL.
 * @param[in]  p_rcv    The ::NTP_PACKET_INFO containing the packet received from the server, in host byte order, or NULL.
 * @param[in]  p_stats  A ::QUERY_STATS structure to be printed, or NULL.
 * @param[in]  print_date_time  Controls if ::print_ntp_packet also prints broken down date and time
 * @param[in]  verbose          Also prints a line with auth result if auth is not used at all
 *
 * @see ::eval_ntp_packet
 */
 void print_ntp_results( const NTP_RESULTS *p_rslt, const NTP_PACKET_INFO *p_req, const NTP_PACKET_INFO *p_rcv, const QUERY_STATS *p_stats, int print_date_time, int verbose ) ;


/* ----- function prototypes end ----- */

#ifdef __cplusplus
}
#endif




#if defined( _USING_BYTE_ALIGNMENT )
  #pragma pack()      // set default alignment
  #undef _USING_BYTE_ALIGNMENT
#endif

/* End of header body */


#undef _ext
#undef _DO_INIT

#endif  /* _MBG_NTP_TEST_UTIL_H */

