
/**************************************************************************
 *
 *  $Id: mbgntp.c 1.2 2018/04/09 07:30:22Z martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Functions useful for NTP implementations.
 *
 * -----------------------------------------------------------------------
 *  $Log: mbgntp.c $
 *  Revision 1.2  2018/04/09 07:30:22Z  martin
 *  Fixed a compiler warning.
 *  Revision 1.1  2018/03/09 09:56:35Z  martin
 *  Initial revision.
 *
 **************************************************************************/

#define _MBGNTP
  #include <mbgntp.h>
#undef _MBGNTP

#include <str_util.h>



/*HDR*/
/**
 * @brief Print the value of the 'precision' field from an NTP packet to a string buffer
 *
 * @param[out]  s         Address of a string buffer to be filled.
 * @param[in]   max_len   Size of the string buffer.
 * @param[in] prec_exp    The precision value from the NTP packet.
 */
int snprint_ntp_prec( char *s, size_t max_len, int8_t prec_exp )
{
  typedef struct
  {
    double limit;
    double scale;
    int frac_digits;
    const char *units;
  } SC_INFO;

  const SC_INFO *p;

  int shift_cnt = abs( prec_exp );
  double d = (double) ( (uint64_t) 1UL << shift_cnt );

  if ( prec_exp < 0 )
  {
    int i;

    #define N_SC_INFO_NEG  7

    static const SC_INFO sc_info_neg[N_SC_INFO_NEG] =
    {
      { 1., 1e3, 0, "ms" },         // 2^^0
      { 8., 1e3, 2, "ms" },         // 2^^3
      { 1024., 1e6, 0, "us" },      // 2^^10
      { 8192., 1e6, 2, "us" },      // 2^^12
      { 524288., 1e9, 0, "ns" },    // 2^^19
      { 8388608., 1e9, 2, "ns" },   // 2^^23
      { 134217728., 1e9, 3, "ns" }  // 2^^27
    };

    for ( i = N_SC_INFO_NEG; i > 0; )
    {
      i--;
      p = &sc_info_neg[i];
      if ( d > p->limit )
        break;
    }

    d = p->scale / d;
  }
  else
  {
    static const SC_INFO sc_info_pos = { 0, 0, 0, "s" };
    p = &sc_info_pos;
  }

  return snprintf_safe( s, max_len, "%.*f %s", p->frac_digits, d, p->units );

}  // snprint_ntp_prec



/*HDR*/
/**
 * @brief Print an NTP refid as text to a string buffer
 *
 * The function ::ntp_refid_should_be_text can be called to determine
 * if the refid is to be interpreted as printable text, so this function
 * can be used to print it into a string buffer.
 *
 * Please note the 4 char refid may not be 0-terminated, so you can't
 * simply cast the original refid to a string, and then print it.
 *
 * @param[out]  s         Address of a string buffer to be filled.
 * @param[in]   max_len   Size of the string buffer.
 * @param[in]   p         Pointer to a ::NTP_REFID.
 *
 * @return Number of characters copied to the destination buffer, without the terminating 0
 *
 * @see ::ntp_refid_should_be_text
 * @see ::snprintf_ntp_refid_as_dotted_quad
 */
int snprintf_ntp_refid_as_text( char *s, size_t max_len, const NTP_REFID *p )
{
  char *cp = s;
  size_t i;

  for ( i = 0; i < sizeof( *p ) && i < max_len; i++ )
  {
    char c = p->as_chars[i];
    *cp++ = ( c > 0x1f && c < 0x7F ) ? c : '.';
  }

  *cp = 0;  // force terminating 0

  return cp - s;

}  // snprintf_ntp_refid_as_text



/*HDR*/
/**
 * @brief Print an NTP refid as dotted quad to a string buffer
 *
 * The function ::ntp_refid_should_be_text can be called to determine
 * if the refid is to be interpreted as printable text, in which case
 * ::snprintf_ntp_refid_as_text should be used to print it, or as binary
 * number, in which case this function should be called to print it.
 *
 * Historically, a binary refid was always the IPv4 address of the upstream
 * time source, so it is usually displayed as dotted quad string.
 * Actually the upstream IP address can also be an IPv6 address which
 * doesn't fit into the 4 byte refid field, so in case of an IPv6 address
 * the binary refid are the 4 LSBs of the MD5 hash of the IPv6 address.
 *
 * @param[out]  s         Address of a string buffer to be filled.
 * @param[in]   max_len   Size of the string buffer.
 * @param[in]   p         Pointer to a ::NTP_REFID.
 *
 * @return Number of characters copied to the destination buffer, without the terminating 0
 *
 * @see ::ntp_refid_should_be_text
 * @see ::snprintf_ntp_refid_as_dotted_quad
 */
int snprintf_ntp_refid_as_dotted_quad( char *s, size_t max_len, const NTP_REFID *p )
{
  int n = 0;
  int i;

  for ( i = 0; ; )
  {
    n += snprintf_safe( &s[n], max_len - n, "%u", p->as_u8s[i] );

    if ( ++i >= sizeof( *p ) )
      break;

    n += sn_cpy_str_safe( &s[n], max_len - n, "." );
  }

  return n;

}  // snprintf_ntp_refid_as_dotted_quad


