
/**************************************************************************
 *
 *  $Id: timeutil.h 1.12 2019/08/07 09:05:44Z martin REL_M $
 *
 *  Copyright (c) Meinberg Funkuhren, Bad Pyrmont, Germany
 *
 *  Description:
 *    Definitions and prototypes for timeutil.c
 *
 * -----------------------------------------------------------------------
 *  $Log: timeutil.h $
 *  Revision 1.12  2019/08/07 09:05:44Z  martin
 *  New inline function mbg_gmtime64().
 *  Updated doxygen comments.
 *  Revision 1.11  2019/07/19 14:29:07  martin
 *  Define CLOCK_MONOTONIC, clock_gettime() and
 *  clock_settime() on Windows, if appropriate.
 *  Updated function prototypes.
 *  Revision 1.10  2019/02/11 09:49:46  martin
 *  Support the mingw build environment.
 *  Fixed build for targets that don't support 64 bit types.
 *  Revision 1.9  2018/12/11 15:36:40  martin
 *  cvt_to_time_t() now expects an int64_t type parameter.
 *  Revision 1.8  2018/11/29 15:32:04Z  martin
 *  Moved some inline functions here.
 *  Include mbgtime.h.
 *  Revision 1.7  2018/01/30 09:54:12Z  martin
 *  Updated function prototypes.
 *  Revision 1.6  2018/01/15 18:31:21Z  martin
 *  Updated function prototypes.
 *  Revision 1.5  2017/11/16 13:34:09  philipp
 *  Extended ISO printing helper functions
 *  Revision 1.4  2017/11/16 11:33:28  philipp
 *  Added functions to print a time_t to ISO format
 *  Revision 1.3  2017/07/05 07:15:00  martin
 *  Provide basic support for clock_gettime()/clock_settime()
 *  compatible functions for Windows.
 *  Updated function prototypes.
 *  Revision 1.1  2016/07/15 14:14:20Z  martin
 *  Initial revision
 *
 **************************************************************************/

#ifndef _TIMEUTIL_H
#define _TIMEUTIL_H

/* Other headers to be included */

#include <gpsdefs.h>
#include <mbgerror.h>
#include <mbgtime.h>

#include <time.h>
#include <stddef.h>


#ifdef _TIMEUTIL
 #define _ext
 #define _DO_INIT
#else
 #define _ext extern
#endif


/* Start of header body */

#ifdef __cplusplus
extern "C" {
#endif


// exclude for MinGW
#if defined( MBG_TGT_WIN32 ) || defined( MBG_TGT_DOS )
  #if !defined( MBG_TGT_MINGW )
    typedef int clockid_t;
    #define clockid_t clockid_t

    #define CLOCK_REALTIME  ( (clockid_t) 0 )
    #define CLOCK_MONOTONIC ( (clockid_t) 1 )

    #define clock_gettime  mbg_clock_gettime
    #define clock_settime  mbg_clock_settime
  #endif
#endif



#if defined( MBG_TGT_WIN32 )

#define __const__ const

/**
 * @brief A pointer to a function returning the system time as FILETIME
 *
 * This can be e.g. the standard Windows API call GetSystemTimeAsFileTime()
 * or the GetSystemTimeAsPreciseFileTime() API call introduced with Windows 8.
 */
typedef VOID (WINAPI *GSTAFT_FNC)(LPFILETIME lpSystemTimeAsFileTime);

_ext GSTAFT_FNC gstaft_fnc
#ifdef _DO_INIT
 = GetSystemTimeAsFileTime
#endif
;

#endif



#if !defined( MBG_TGT_MISSING_64_BIT_TYPES )
  typedef int64_t mbg_time_t;   // we try to always use 64 bit types
#else
  typedef time_t mbg_time_t;    // fall back to the default
#endif


static __mbg_inline
time_t cvt_to_time_t( mbg_time_t t )
{
  // Eventually we can do some epoch check / conversion here.
  return (time_t) t;

}  // cvt_to_time_t



static __mbg_inline
/**
 * @brief A replacement function for POSIX gmtime().
 *
 * This function calls the original gmtime() function, but unlike gmtime()
 * it expects the address of a struct tm variable which is only filled
 * if gmtime() completed successfully. An appropriate return code
 * is provided, indicating if the conversion succeeded, or not.
 *
 * The original gmtime() function returns a NULL pointer if the conversion
 * fails, so a programs which just uses the pointer without checking it first
 * may trap if the conversion fails. This can't happen with this function.
 *
 * This variant expects the address of a generic time_t variable. A time_t can
 * be 32 bit or 64 bit wide, depending on the build environment and target system.
 * Systems with 32 bit time_t will suffer from the Y2038 problem.
 *
 * There is also the ::mbg_gmtime64 variant, which expects a pointer to an
 * ::MBG_TIME64_T variable to be converted, which is always 64 bits, even
 * on 32 bit systems.
 *
 * @param[out]  p_tm  Address of a struct tm variable to take the conversion result.
 * @param[in]   p_t   Address of a time_t variable to be converted.
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 *
 * @see ::mbg_gmtime64
 */
int mbg_gmtime( struct tm *p_tm, const time_t *p_t )
{
  struct tm *p_tm_tmp = gmtime( p_t );

  if ( p_tm_tmp == NULL )  // Conversion failed.
    return mbg_get_last_error( NULL );

  *p_tm = *p_tm_tmp;

  return MBG_SUCCESS;

}  // mbg_gmtime



static __mbg_inline
/**
 * @brief A replacement function for POSIX gmtime().
 *
 * This variant of ::mbg_gmtime expects a pointer to an ::MBG_TIME64_T variable
 * to be converted, which is always 64 bits, even on 32 bit systems.
 *
 * Actually, the ::MBG_TIME64_T value is truncated to a native time_t, which can
 * be 32 bit or 64 bit wide, depending on the build environment and target system.
 * So this "just works" on systems with 64 bit time_t, but systems with 32 bit time_t
 * will anyway suffer from the Y2038 problem, unless the gmtime() call is replaced
 * by a function that does a 64 bit conversion even on systems with 32 bit time_t.
 *
 * @param[out]  p_tm  Address of a struct tm variable to take the conversion result.
 * @param[in]   p_t   Address of an ::MBG_TIME64_T variable to be converted.
 *
 * @return ::MBG_SUCCESS on success, else one of the @ref MBG_ERROR_CODES
 *
 * @see ::mbg_gmtime
 */
int mbg_gmtime64( struct tm *p_tm, const MBG_TIME64_T *p_t )
{
  struct tm *p_tm_tmp;
  time_t t = (time_t) (*p_t);

  p_tm_tmp = gmtime( &t );

  if ( p_tm_tmp == NULL )  // Conversion failed.
    return mbg_get_last_error( NULL );

  *p_tm = *p_tm_tmp;

  return MBG_SUCCESS;

}  // mbg_gmtime64



#if !defined( MBG_TGT_MISSING_64_BIT_TYPES )

static __mbg_inline /*HDR*/
double ntp_tstamp_to_double( const NTP_TSTAMP *t )
{
  return (double) t->seconds + ( ( (double) t->fractions ) / NTP_FRAC_PER_SEC );

}  // ntp_tstamp_to_double



static __mbg_inline /*HDR*/
ulong ntp_frac_to_nsec( uint32_t frac )
{
  uint64_t tmp = ( (uint64_t) frac * NSEC_PER_SEC ) / NTP_FRAC_PER_SEC;

#if 0  // TODO Currently not supported
  if ( tmp >= NSEC_PER_SEC )
    mbglog( LOG_WARNING, "Range overflow in ntp_frac_to_nsec: 0x%X -> %Lu",
            frac, tmp );
#endif

  return (ulong) tmp;

}  // ntp_frac_to_nsec



static __mbg_inline /*HDR*/
uint32_t nsec_to_ntp_frac( ulong nsec )
{
  uint64_t tmp;

  tmp = ( (uint64_t) nsec * NTP_FRAC_PER_SEC ) / NSEC_PER_SEC;

#if 0  // TODO Currently not supported
  if ( tmp >= NTP_FRAC_PER_SEC )
    mbglog( LOG_WARNING, "Range overflow in nsec_to_ntp_frac: %lu -> 0x%LX",
            (ulong) nsec, tmp );
#endif

  return (uint32_t) tmp;

}  // nsec_to_ntp_frac



static __mbg_inline /*HDR*/
void timespec_to_ntp_tstamp( NTP_TSTAMP *t_ntp, const struct timespec *t_ts )
{
  t_ntp->seconds = (uint32_t) t_ts->tv_sec + NTP_SEC_BIAS;
  t_ntp->fractions = nsec_to_ntp_frac( t_ts->tv_nsec );

}  // timespec_to_ntp_tstamp



static __mbg_inline /*HDR*/
void ntp_tstamp_to_timespec( struct timespec *t_ts, const NTP_TSTAMP *t_ntp )
{
  t_ts->tv_sec = t_ntp->seconds - NTP_SEC_BIAS;
  t_ts->tv_nsec = ntp_frac_to_nsec( t_ntp->fractions );

}  // ntp_tstamp_to_timespec

#endif  // !defined( MBG_TGT_MISSING_64_BIT_TYPES )



static __mbg_inline /*HDR*/
int timespec_is_set( const struct timespec *p )
{
  return p->tv_sec != 0 || p->tv_nsec != 0;

}  // timespec_is_set



//
// compute the difference of two timespec variables
//
static __mbg_inline /*HDR*/
double delta_timespec_d_s( const struct timespec *ts,
                           const struct timespec *ts_ref )
{
  return ( (double) ts->tv_sec - (double) ts_ref->tv_sec )
       + ( (double) ts->tv_nsec - (double) ts_ref->tv_nsec ) / NSEC_PER_SEC;

}  // delta_timespec_d_s



#if !defined( MBG_TGT_MISSING_64_BIT_TYPES )

static __mbg_inline /*HDR*/
int64_t delta_timespec_ll_ns( const struct timespec *ts,
                              const struct timespec *ts_ref )
{
  int64_t tmp = ts->tv_sec - ts_ref->tv_sec;
  tmp = ( tmp * NSEC_PER_SEC ) + ( ts->tv_nsec - ts_ref->tv_nsec );

  return tmp;

}  // delta_timespec_ll_ns

#endif  // !defined( MBG_TGT_MISSING_64_BIT_TYPES )



/* ----- function prototypes begin ----- */

/* This section was generated automatically */
/* by MAKEHDR, do not remove the comments. */

 int snprint_gmtime_error( char *s, size_t max_len, int mbg_errno, time_t t, const char *calling_fnc ) ;
 /**
 * @brief A Windows implementation for POSIX clock_gettime().
 *
 * @param[in]   clock_id  Identifier of a specific clock, e.g.
 *                        CLOCK_REALTIME, CLOCK_MONOTONIC, etc.
 * @param[out]  tp        Address of a struct timespec to take up
 *                        the current time.
 *
 * @return  0 on success, -1 on error, just like the POSIX function.
 *          In case of an error the POSIX errno variable is set
 *          appropriately.
 */
 int mbg_clock_gettime( clockid_t clock_id, struct timespec *tp ) ;

 /**
 * @brief A Windows implementation for POSIX clock_settime().
 *
 * @param[in]  clock_id  Identifier of a specific clock, i.e.
 *                       CLOCK_REALTIME which is the only clock
 *                       supported by this call.
 * @param[in]  tp        Pointer to a struct timespec providing
 *                       the time to be set.
 *
 * @return  0 on success, -1 on error, just like the POSIX function.
 *          In case of an error the POSIX errno variable is set
 *          appropriately.
 */
 int mbg_clock_settime( clockid_t clock_id, const struct timespec *tp ) ;

 void check_precise_time_api( void ) ;
 /**
 * @brief Print a UTC offset into a string
 *
 * Format of the string is "[info]+hh[:mm[:ss]]h"
 *
 * @param[out]  s         Address of a string buffer to be filled.
 * @param[in]   max_len   Size of the string buffer.
 * @param[in]   info      An optional info string to be prepended, may be NULL.
 * @param[in]   utc_offs  The UTC offset to be printed, in [s].
 *
 * @return The number of characters written to the output buffer, except the terminating 0.
 */
 int snprint_utc_offs( char *s, size_t max_len, const char *info, long utc_offs ) ;

 int snprint_time_t_to_iso( time_t tstamp, int offs_hours, char *buf, size_t len ) ;
 int snprint_time_t_to_iso_ms( time_t tstamp, int offs_hours, time_t frac, char *buf, size_t len ) ;
 int snprint_time_t_to_iso_us( time_t tstamp, int offs_hours, time_t frac, char *buf, size_t len ) ;
 int snprint_time_t_to_iso_ns( time_t tstamp, int offs_hours, time_t frac, char *buf, size_t len ) ;
 int snprint_ntp_tstamp_to_iso_ms( const NTP_TSTAMP *ts, char *buf, size_t len ) ;
 int snprint_ntp_tstamp_to_iso_us( const NTP_TSTAMP *ts, char *buf, size_t len ) ;
 int snprint_ntp_tstamp_to_iso_ns( const NTP_TSTAMP *ts, char *buf, size_t len ) ;

/* ----- function prototypes end ----- */

#ifdef __cplusplus
}
#endif


/* End of header body */

#undef _ext
#undef _DO_INIT

#endif  /* _TIMEUTIL_H */
